import random

from .gamestate import GameStateView


class BaseAI:

    def __init__(self, game_state_view: GameStateView):
        self.game_state_view = game_state_view

    def pick_action(self, actions):
        raise NotImplementedError()


class RandomAI(BaseAI):

    def pick_action(self, actions):
        action, = random.sample(actions, 1)
        return action



from dataclasses import dataclass

from .cards import Trick
from .player import Player
from .partners import TeamID
from .game_actions import BaseAction


class BaseEvent:
    pass


@dataclass(frozen=True)
class ActionTakenEvent(BaseEvent):
    player: Player
    action: BaseAction

    def __str__(self):
        return f"Player {self.player.name}: {self.action}"


@dataclass(frozen=True)
class TrickTakenEvent(BaseEvent):
    player: Player
    team_id: TeamID
    trick: Trick

    def __str__(self):
        trick_symbols = tuple(card.symbol for card in self.trick)
        return f"Player {self.player.name} took {trick_symbols}"



import random
from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Tuple


# --- OrderedEnum ---


class OrderedEnum(Enum):
    """
    See https://docs.python.org/3/library/enum.html#orderedenum
    """

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


# --- Suit ---


class Suit(OrderedEnum):
    Unknown = auto()
    Club = auto()
    Diamond = auto()
    Heart = auto()
    Spade = auto()

    @property
    def symbol(self):
        return suit_symbol[self]


suit_symbol = {
    Suit.Unknown: "_",
    Suit.Club: "♣",
    Suit.Diamond: "♦",
    Suit.Heart: "♥",
    Suit.Spade: "♠",
}

suits = (
    Suit.Club,
    Suit.Diamond,
    Suit.Heart,
    Suit.Spade,
)


# --- Rank ---


class Rank(OrderedEnum):
    Unknown = auto()
    Two = auto()
    Three = auto()
    Four = auto()
    Five = auto()
    Six = auto()
    Seven = auto()
    Eight = auto()
    Nine = auto()
    Ten = auto()
    Jack = auto()
    Queen = auto()
    King = auto()
    Ace = auto()
    Joker = auto()

    @property
    def symbol(self):
        return rank_symbol[self]


rank_symbol = {
    Rank.Unknown: "_",
    Rank.Ace: "A",
    Rank.Two: "2",
    Rank.Three: "3",
    Rank.Four: "4",
    Rank.Five: "5",
    Rank.Six: "6",
    Rank.Seven: "7",
    Rank.Eight: "8",
    Rank.Nine: "9",
    Rank.Ten: "10",
    Rank.Jack: "J",
    Rank.Queen: "Q",
    Rank.King: "K",
    Rank.Joker: "🃏",
}

ranks = (
    Rank.Ace, Rank.Two, Rank.Three,
    Rank.Four, Rank.Five, Rank.Six,
    Rank.Seven, Rank.Eight, Rank.Nine, Rank.Ten,
    Rank.Jack, Rank.Queen, Rank.King,
)


# --- Card ----


@dataclass(order=True, frozen=True)
class Card:
    suit: Suit
    rank: Rank

    @property
    def symbol(self):
        return card_symbol[self]


Card.unknown = Card(Suit.Unknown, Rank.Unknown)
Card.joker = Card(Suit.Unknown, Rank.Joker)


card_symbol = {
    Card.unknown: "🂠",
    Card.joker: "🃏",
}

card_symbol_unicode = {
    Suit.Heart: "🂱 🂲 🂳 🂴 🂵 🂶 🂷 🂸 🂹 🂺 🂻 🂽 🂾",
    Suit.Spade: "🂡 🂢 🂣 🂤 🂥 🂦 🂧 🂨 🂩 🂪 🂫 🂭 🂮",
    Suit.Diamond: "🃁 🃂 🃃 🃄 🃅 🃆 🃇 🃈 🃉 🃊 🃋 🃍 🃎",
    Suit.Club: "🃑 🃒 🃓 🃔 🃕 🃖 🃗 🃘 🃙 🃚 🃛 🃝 🃞",
}

for suit, line in card_symbol_unicode.items():
    symbols = line.split(" ")
    for rank, symbol in zip(ranks, symbols):
        card = Card(suit, rank)
        card_symbol[card] = symbol


# --- Deck ---


@dataclass
class Deck:
    cards: List[Card]
    allow_reorder: bool = True

    @staticmethod
    def empty():
        return Deck([])

    @staticmethod
    def empty_pile():
        deck = Deck.empty()
        deck.allow_reorder = False
        return deck

    @staticmethod
    def from_deck(other: "Deck"):
        return Deck(other.cards.copy())

    @staticmethod
    def full_deck(num_jokers=3):
        deck = Deck.empty()

        for suit in suits:
            for rank in ranks:
                deck.cards.append(Card(suit, rank))

        for _ in range(num_jokers):
            deck.cards.append(Card.joker)

        return deck

    def sort(self, trump=Suit.Unknown):
        if self.allow_reorder:
            self.cards.sort()

    def shuffle(self):
        if self.allow_reorder:
            random.shuffle(self.cards)

    def take(self, card: Card):
        return self.cards.remove(card)

    def give(self, card: Card, sort=True):
        self.cards.append(card)
        if sort:
            self.sort()

    def to_trick(self) -> "Trick":
        if len(self) != 4:
            raise ValueError("to_trick only support tricks of size 4")

        return tuple(self.cards)

    def __len__(self):
        return len(self.cards)

    def __iter__(self):
        return iter(self.cards)

    def __str__(self):
        return " ".join(c.symbol for c in self.cards)


# --- Trick ---


Trick4 = Tuple[Card, Card, Card, Card]
Trick = Trick4



from dataclasses import dataclass

from .player import Player
from .cards import Card


class BaseAction:
    """
    Base class for game actions
    """


@dataclass(frozen=True)
class PlayAction(BaseAction):
    card: Card

    def __str__(self):
        return f"Played {self.card.symbol}"



from dataclasses import dataclass
from collections import defaultdict
from typing import List, Dict

from .player import Player


Team = List[Player]
TeamID = int


@dataclass
class Partners:
    players: List[Player]
    team_id_assignment: Dict[Player, TeamID] = None

    def __post_init__(self):
        if self.team_id_assignment is None:
            # Give each player their own team
            self.team_id_assignment = {
                player: self.initial_team_assignment(i)
                for i, player in enumerate(self.players)
            }

    def initial_team_assignment(self, index):
        offset = 1000
        return index + offset

    def join(self, *players):
        if len(players) == 0:
            return

        if len(players) > len(self.players):
            raise ValueError("Cannot join more players than there exists.")

        first_player = players[0]
        team_id = self.team_id(first_player)

        for player in players:
            self.team_id_assignment[player] = team_id

    def isolate(self, player):
        rest_of_players = [
            p for p in self.players
            if p != player
        ]
        self.join(*rest_of_players)

    def bisect(self, *team_1_players):
        team_2_players = [
            p for p in self.players
            if p not in team_1_players
        ]
        self.join(*team_1_players)
        self.join(*team_2_players)

    def team_id(self, player: Player):
        return self.team_id_assignment[player]

    def team_members(self, team_id: TeamID) -> Team:
        return [
            player
            for player in self.players
            if self.team_id(player) == team_id
        ]

    def team_size(self, team_id: TeamID):
        return len(self.team_members(team_id))

    @property
    def teams(self):
        d = defaultdict(list)
        for player in self.players:
            team_id = self.team_id(player)
            d[team_id].add(player)
        return dict(d)

    @property
    def team_ids(self):
        return set(
            self.team_id(player)
            for player in self.players
        )

    @property
    def num_teams(self):
        return len(self.team_ids)



from dataclasses import dataclass

from .cards import Suit, Card, Deck
from .gamestate import GameState, GameStateView, Player, Partners
from .game_actions import BaseAction, PlayAction
from .game_events import ActionTakenEvent, TrickTakenEvent
from .game_ai import RandomAI


@dataclass
class Game:
    state: GameState = None

    def __post_init__(self):
        if self.state is None:
            self.state = self.initial_state()

    @staticmethod
    def initial_state(**settings):
        num_players = settings.get("num_players", 4)
        player_names = ("north", "east", "south", "west")

        # Create players (up to 4) based on preset player names
        players = [
            Player(id, name)
            for id, (_, name)
            in enumerate(zip(range(num_players), player_names))
        ]

        # Initially each player starts with an empty hand
        hands = {
            p: Deck.empty()
            for p in players
        }

        # Initally, we don't know who are partners.
        partners = Partners(players)
        partners.bisect(players[0], players[2])

        return GameState(
            players=players,
            hands=hands,
            partners=partners,
        )

    def deal(self):
        deck = Deck.full_deck()
        deck.shuffle()

        hand_size = 13
        assert len(self.state.players) == 4

        for p in self.state.players:
            hand = self.state.hands[p]
            for _ in range(hand_size):
                card = deck.cards.pop()
                hand.give(card, sort=False)
            hand.sort()

        self.state.kitty = Deck.from_deck(deck)

    @property
    def has_ended(self):
        hand_size = 13
        return len(self.state.tricks) >= hand_size

    @property
    def current_player(self):
        return self.state.current_player

    def valid_actions(self, player):
        if player != self.current_player:
            return []

        hand = self.state.hands[player]
        pile = self.state.pile

        if len(pile) == 0:
            return [PlayAction(card) for card in hand]
        else:
            first_card = pile.cards[0]

            if first_card == Card.joker:
                suits_from_hand = hand
            else:
                suits_from_hand = [
                    card for card in hand
                    if card.suit is first_card.suit
                ]

            if len(suits_from_hand) > 0:
                return [PlayAction(card) for card in suits_from_hand]
            else:  # Renons
                return [PlayAction(card) for card in hand]

        return []

    def is_valid_action(self, player, action: PlayAction):
        if player != self.current_player:
            return False

        return action in self.valid_actions(player)

    def take_action(self, player, action):
        if not isinstance(action, BaseAction):
            raise ValueError(f"Not an action: {action}")

        self.state.events.append(ActionTakenEvent(player, action))

        if isinstance(action, PlayAction):
            state = self.state
            hand = state.hands[player]
            hand.take(action.card)

            pile = state.pile

            if len(pile) == 0:
                state.pile_suit = action.card.suit

            pile.give(action.card)
            state.pile_play.append(player)

            if len(pile) == state.num_players:
                # Score trick
                card_scores = self._score_pile(pile)

                # Determine winning team
                trick_winner = self._assign_trick(self.state.pile_play, card_scores)
                team_id = self.state.partners.team_id(trick_winner)

                # Add trick to list of tricks
                trick = pile.to_trick()
                trick_index = len(state.tricks)
                state.tricks.append(trick)
                state.trick_owner[trick_index] = team_id

                # Reset round
                state.round_reset()

                self.state.events.append(TrickTakenEvent(trick_winner, team_id, trick))

                state.turn = state.players.index(trick_winner)

            else:
                state.turn = (state.turn + 1) % state.num_players
        else:
            raise ValueError(f"Invalid action: {action}")

        return True

    def _score_pile(self, pile) -> list:
        if len(pile) == 0:
            return {}

        first_card = pile.cards[0]
        valid_trump = self.state.trump is not Suit.Unknown

        card_scores = []
        for card in pile:
            value = card.rank.value
            if valid_trump and card.suit is self.state.trump:
                value += 100
            elif card.suit is not first_card.suit:
                value -= 100
            card_scores.append(value)

        if first_card == Card.joker:
            card_scores[0] += 1000

        return card_scores

    def _assign_trick(self, pile_play, scores):
        _, winner = max(zip(scores, pile_play))
        return winner




# --- CLI ---


def display_board(view):
    for p, hand in view.other_players.items():
        print(f"{p.name:>6}: {hand}")

    print()
    for p in view.partners:
        print(f"Your partner: {p.name}")

    print()
    print(f"Pile: {view.pile}")
    print()
    print(f"Your hand: {view.hand}")
    print()



def parse_cli_action(s, actions):
    try:
        index = int(s)
    except ValueError:
        return None

    try:
        return actions[index]
    except IndexError:
        return None


def display_actions(actions):
    for i, action in enumerate(actions):
        if isinstance(action, PlayAction):
            s = action.card.symbol
        else:
            s = "?"
        print(f"{i:>2}: {s}")


if __name__ == "__main__":
    import time

    # Interactive CLI game


    last_event_index = 0

    game = Game()

    controllers = []
    for player in game.state.players:
        view = GameStateView(game.state, player)
        ai = RandomAI(view)
        controllers.append(ai)

    my_view = controllers[0].game_state_view
    controllers[0] = "human"

    game.deal()

    while not game.has_ended:
        player = game.current_player
        controller = controllers[game.state.turn]

        if controller == "human":
            if len(game.state.events) > last_event_index:
                print()
                print("=== Events: ===")
                for event in game.state.events[last_event_index:]:
                    print("-", event)
                last_event_index = len(game.state.events)
                print()

            print()
            print(f"=== Turn: {player.name} ===")
            print()

            display_board(my_view)

            print("Actions:")
            actions = game.valid_actions(player)
            display_actions(actions)

            while True:
                s = input("> ")
                if s == "tricks":
                    print(game.state.tricks)
                    print(game.state.trick_owner)
                    continue

                action = parse_cli_action(s, actions)
                if game.is_valid_action(player, action):
                    result = game.take_action(player, action)
                    if result:
                        break
                else:
                    print("Invalid action.")
        else:
            actions = game.valid_actions(player)
            action = controller.pick_action(actions)
            game.take_action(player, action)

            print()
            print(f"=== Turn taken: {player.name} ===")
            print()

            display_board(my_view)
            time.sleep(1)


    print("Tricks:")

    from collections import Counter

    score = Counter()

    for trick_index, team_id in game.state.trick_owner.items():
        score[team_id] += 1

    for team_id, s in score.items():
        members = game.state.partners.team_members(team_id)
        team_str = ", ".join(p.name for p in members)
        print(f"{team_str:<13}: {s}")


# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()

setup(
    name="whist",
    version="0.1.0",
    description="A package containing the logic for the card game Whist and its variants.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Peter Severin Rasmussen",
    author_email="git@petersr.com",
    url="https://github.com/PeterSR/pywhist",
    license="MIT License",
    packages=find_packages(exclude=("tests", "docs")),
    python_requires=">=3.7",
)



from dataclasses import dataclass
from typing import List, Dict

from .cards import Suit, Card, Deck, Trick
from .player import Player
from .partners import Partners, TeamID
from .game_events import BaseEvent


@dataclass
class GameState:
    players: List[Player] = None
    hands: Dict[Player, Deck] = None
    trump: Suit = Suit.Unknown
    partners: Partners = None
    kitty: Deck = None
    pile: Deck = None
    pile_play: List[Player] = None
    tricks: List[Trick] = None
    trick_owner: Dict[int, TeamID] = None

    events: List[BaseEvent] = None

    # Index in players
    turn: int = 0

    bid: str = ""

    mode: str = ""

    def __post_init__(self):
        self.round_reset()
        self.tricks = []
        self.trick_owner = {}
        self.events = []

        if self.partners is None:
            self.partners = Partners(self.players)

    @property
    def num_players(self):
        return len(self.players)

    @property
    def current_player(self):
        return self.players[self.turn]

    def round_reset(self):
        self.pile = Deck.empty_pile()
        self.pile_play = []


class GameStateView:
    """
    Represents one players view of the current state of the game.
    """

    def __init__(self, state, player):
        self.state = state
        self.player = player

    @property
    def pile(self):
        return self.state.pile

    @property
    def hand(self):
        return self.state.hands[self.player]

    @property
    def other_players(self):
        players = tuple(p for p in self.state.players if p is not self.player)

        hands = {}

        for p in players:
            num_cards = len(self.state.hands[p].cards)
            hands[p] = Deck([Card.unknown] * num_cards)

        return hands

    @property
    def partners(self):
        team_id = self.state.partners.team_id(self.player)
        members = self.state.partners.team_members(team_id)
        members.remove(self.player)
        return tuple(members)
